import json
import os


class Tokenizer:
    """
    tokenizer class will tokenize the input interms of charater level 
    based on the the given text_data to get_encoding function it will use vacobalory
    if not it will use the default encodings to encode and decode data
    """

    def __init__(self):
        pass

    def get_encoding(self, model: str = None):

        self.vocab = Tokenizer.get_char_vocab(f"{model}/vocab.json")
        self.stoi, self.itos = self.vocab
        self.n_vocab = len(self.stoi)
        # print(self.vocab))
        # print(self.n_vocab)

    def set_encoding(self, model: str, data: str):
        """
        toake text or string data and segregate it into vocab and store it in model
        """
        # handelling folder not existerror
        if not os.path.exists(f"./{model}"):
            os.makedirs(f"./{model}")
        Tokenizer.set_char_vocab(f"{model}/vocab.json", data)

    def encode(self, s: str) -> list[int]:
        # encoder: take a string char , output a list of integers
        enc_list = []
        for c in s:
            if c not in self.stoi:
                self.stoi[c] = max(self.stoi.values())+1
                self.n_vocab += 1
            enc_list.append(self.stoi[c])
        return enc_list

    def decode(self, l: list[int]) -> str:
        # decoder: take a list of integers, output a string
        return ''.join([self.itos[str(i)] for i in l])

    @classmethod
    def get_char_vocab(cls, path: str):
        """
        json file -> dict -> dict,dict
        """
        with open(path, "r", encoding="utf-8") as f:
            data = json.load(f)
            stoi = data["stoi"]
            itos = data["itos"]
        return stoi, itos

    @classmethod
    def set_char_vocab(cls, path: str, data: str) -> None:
        """
        string data -> vocab -> dict,dict -> dict -> json file
        """
        data_chars = sorted(list(set(data)))
        stoi = {ch: i for i, ch in enumerate(data_chars)}
        itos = {i: ch for i, ch in enumerate(data_chars)}
        vocab = {"stoi": stoi, "itos": itos}
        with open(path, "w", encoding="utf-8") as f:
            json.dump(vocab, f)
            # print("writen data string : ",len(data_string))



import torch
import torch.nn as nn
from torch.nn import functional as F
from . import config

torch.manual_seed(1337)


"""
defines a language model using PyTorch 
based on the Transformer architecture attention Mechanism
"""
[vocab_size,
 batch_size,
 block_size,
 max_iters,
 eval_interval,
 learning_rate,
 device,
 eval_iters,
 n_embd,
 n_head,
 n_layer,
 dropout] = config.pass_args()


def set_parameters(args):
    global vocab_size, batch_size, block_size, max_iters, eval_interval, learning_rate, device, eval_iters, n_embd, n_head, n_layer, dropout
    [vocab_size, batch_size, block_size, max_iters, eval_interval,
        learning_rate, device, eval_iters, n_embd, n_head, n_layer, dropout] = args


class Head(nn.Module):
    """ one head of self-attention """

    def __init__(self, head_size):
        super().__init__()
        self.key = nn.Linear(n_embd, head_size, bias=False)
        self.query = nn.Linear(n_embd, head_size, bias=False)
        self.value = nn.Linear(n_embd, head_size, bias=False)
        self.register_buffer('tril', torch.tril(
            torch.ones(block_size, block_size)))

        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        B, T, C = x.shape
        k = self.key(x)   # (B,T,C)
        q = self.query(x)  # (B,T,C)
        # compute attention scores ("affinities")
        # (B, T, C) @ (B, C, T) -> (B, T, T)
        wei = q @ k.transpose(-2, -1) * C**-0.5
        wei = wei.masked_fill(
            self.tril[:T, :T] == 0, float('-inf'))  # (B, T, T)
        wei = F.softmax(wei, dim=-1)  # (B, T, T)
        wei = self.dropout(wei)
        # perform the weighted aggregation of the values
        v = self.value(x)  # (B,T,C)
        out = wei @ v  # (B, T, T) @ (B, T, C) -> (B, T, C)
        return out


class MultiHeadAttention(nn.Module):
    """ multiple heads of self-attention in parallel """

    def __init__(self, num_heads, head_size):
        super().__init__()
        self.heads = nn.ModuleList(
            [Head(head_size) for _ in range(num_heads)])
        self.proj = nn.Linear(n_embd, n_embd)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        out = torch.cat([h(x) for h in self.heads], dim=-1)
        out = self.dropout(self.proj(out))
        return out


class FeedFoward(nn.Module):
    """ a simple linear layer followed by a non-linearity """

    def __init__(self, n_embd):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(n_embd, 4 * n_embd),
            nn.ReLU(),
            nn.Linear(4 * n_embd, n_embd),
            nn.Dropout(dropout),
        )

    def forward(self, x):
        return self.net(x)


class Block(nn.Module):
    """ Transformer block: communication followed by computation """

    def __init__(self, n_embd, n_head):
        # n_embd: embedding dimension, n_head: the number of heads we'd like
        super().__init__()
        head_size = n_embd // n_head
        self.sa = MultiHeadAttention(n_head, head_size)
        self.ffwd = FeedFoward(n_embd)
        self.ln1 = nn.LayerNorm(n_embd)
        self.ln2 = nn.LayerNorm(n_embd)

    def forward(self, x):
        x = x + self.sa(self.ln1(x))
        x = x + self.ffwd(self.ln2(x))
        return x

# ---------------------------models-------------------------------

# super simple bigram model


class BigramLM(nn.Module):
    """
    language model predicts next token 
    based on previous tokens using a simple bigram approach
    """

    def __init__(self):
        super().__init__()
        # setting updated parameters
        set_parameters(config.pass_args())

        # each token directly reads off the logits for the next token from a lookup table
        self.token_embedding_table = nn.Embedding(vocab_size, n_embd)
        self.position_embedding_table = nn.Embedding(block_size, n_embd)
        self.blocks = nn.Sequential(
            *[Block(n_embd, n_head=n_head) for _ in range(n_layer)])
        self.ln_f = nn.LayerNorm(n_embd)  # final layer norm
        self.lm_head = nn.Linear(n_embd, vocab_size)

    def forward(self, idx, targets=None):
        B, T = idx.shape

        # idx and targets are both (B,T) tensor of integers
        tok_emb = self.token_embedding_table(idx)  # (B,T,C)
        pos_emb = self.position_embedding_table(
            torch.arange(T, device=device))  # (T,C)
        x = tok_emb + pos_emb  # (B,T,C)
        x = self.blocks(x)  # (B,T,C)
        x = self.ln_f(x)  # (B,T,C)
        logits = self.lm_head(x)  # (B,T,vocab_size)

        if targets is None:
            loss = None
        else:
            B, T, C = logits.shape
            logits = logits.view(B*T, C)
            targets = targets.view(B*T)
            loss = F.cross_entropy(logits, targets)

        return logits, loss

    def generate(self, idx, max_new_tokens):
        # idx is (B, T) array of indices in the current context
        for _ in range(max_new_tokens):
            # crop idx to the last block_size tokens
            idx_cond = idx[:, -block_size:]
            # get the predictions
            logits, loss = self(idx_cond)
            # focus only on the last time step
            logits = logits[:, -1, :]  # becomes (B, C)
            # apply softmax to get probabilities
            probs = F.softmax(logits, dim=-1)  # (B, C)
            # sample from the distribution
            idx_next = torch.multinomial(probs, num_samples=1)  # (B, 1)
            # append sampled index to the running sequence
            idx = torch.cat((idx, idx_next), dim=1)  # (B, T+1)
        return idx


class GPTLM(nn.Module):
    """
    GPT-style approach where each token predicts the next token
    """

    def __init__(self):
        super().__init__()
        # setting updated parameters
        set_parameters(config.pass_args())

        # each token directly reads off the logits for the next token from a lookup table
        self.token_embedding_table = nn.Embedding(vocab_size, n_embd)
        self.position_embedding_table = nn.Embedding(block_size, n_embd)
        self.blocks = nn.Sequential(
            *[Block(n_embd, n_head=n_head) for _ in range(n_layer)])
        self.ln_f = nn.LayerNorm(n_embd)  # final layer norm
        self.lm_head = nn.Linear(n_embd, vocab_size)

        # better init, not covered in the original GPT video, but important, will cover in followup video
        self.apply(self._init_weights)

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

    def forward(self, idx, targets=None):
        B, T = idx.shape

        # idx and targets are both (B,T) tensor of integers
        tok_emb = self.token_embedding_table(idx)  # (B,T,C)
        pos_emb = self.position_embedding_table(
            torch.arange(T, device=device))  # (T,C)
        x = tok_emb + pos_emb  # (B,T,C)
        x = self.blocks(x)  # (B,T,C)
        x = self.ln_f(x)  # (B,T,C)
        logits = self.lm_head(x)  # (B,T,vocab_size)

        if targets is None:
            loss = None
        else:
            B, T, C = logits.shape
            logits = logits.view(B*T, C)
            targets = targets.view(B*T)
            loss = F.cross_entropy(logits, targets)

        return logits, loss

    def generate(self, idx, max_new_tokens):
        # idx is (B, T) array of indices in the current context
        for _ in range(max_new_tokens):
            # crop idx to the last block_size tokens
            # idx_cond = idx[:, -block_size:]
            idx_cond = idx if idx.size(
                1) <= block_size else idx[:, -block_size:]
            # get the predictions
            logits, loss = self(idx_cond)
            # focus only on the last time step
            logits = logits[:, -1, :]  # becomes (B, C)
            # apply softmax to get probabilities
            probs = F.softmax(logits, dim=-1)  # (B, C)
            # sample from the distribution
            idx_next = torch.multinomial(probs, num_samples=1)  # (B, 1)
            # append sampled index to the running sequence
            idx = torch.cat((idx, idx_next), dim=1)  # (B, T+1)
        return idx


class JamGPTLM(nn.Module):

    def __init__(self):
        super().__init__()
        pass



import torch
from . import config


class Data:
    """ just a data preprocessor """
    def __init__(self) -> None:
        pass

    @classmethod
    def get(cls, path: str) -> str:
        """
        text data file -> string data

        """
        with open(path, "r", encoding="utf-8") as f:
            text_data = f.read()
        return text_data

    @classmethod
    def set(cls, path: str, data: str) -> None:
        """
        string data -> text data file
        """
        with open(path, "w", encoding="utf-8") as f:
            f.write(data)
            print("writen data : ", len(data))

    @classmethod
    def train_test_split(cls, data, split_percent: int = 90):
        # split the data into train ans test based on split percentage

        tensor_data = torch.tensor(
            data, dtype=torch.long, device=config.device)
        n = int((split_percent/100)*len(tensor_data))
        train_data = tensor_data[:n]
        test_data = tensor_data[n:]

        return [train_data, test_data]


def test(path):
    Data.set(path, "ghsdubvujsjbnjsnjbnsjnbljnlbnls")
    d = Data.get(path)
    print(d)
    Data.set_vocab(path, d)
    l = Data.get_vocab(path)
    print(l)



import json
import torch


# -----------------------------------------------------------#
# hyperparameters

vocab_size = 0
batch_size = 16
block_size = 32
max_iters = 5000
eval_interval = 100
learning_rate = 1e-3
device = "cuda" if torch.cuda.is_available() else "cpu"
eval_iters = 200
n_embd = 64
n_head = 4
n_layer = 4
dropout = 0.0

# -----------------------------------------------------------#


def pass_args():
    return [vocab_size, batch_size, block_size, max_iters, eval_interval, learning_rate, device, eval_iters, n_embd, n_head, n_layer, dropout]


def get_args():
    args = pass_args()
    arg_names = ['vocab_size', 'batch_size', 'block_size', 'max_iters', 'eval_interval',
                 'learning_rate', 'device', 'eval_iters', 'n_embd', 'n_head', 'n_layer', 'dropout']
    print("# -------------------------------------#\n# hyperparameters")
    max_arg_name_length = max(len(arg) for arg in arg_names)
    for arg_name, arg_value in zip(arg_names, args):
        padding = ' ' * (max_arg_name_length - len(arg_name))
        print(f"{arg_name} {padding} :  {arg_value}")
    print("# -------------------------------------#")
    return args


def set_args(args: list):
    global vocab_size, batch_size, block_size, max_iters, eval_interval, learning_rate, device, eval_iters, n_embd, n_head, n_layer, dropout
    [vocab_size, batch_size, block_size, max_iters, eval_interval,
        learning_rate, device, eval_iters, n_embd, n_head, n_layer, dropout] = args


def variables_to_dict():
    config_dict = {
        "vocab_size": vocab_size,
        "batch_size": batch_size,
        "block_size": block_size,
        "max_iters": max_iters,
        "eval_interval": eval_interval,
        "learning_rate": learning_rate,
        "device": device,
        "eval_iters": eval_iters,
        "n_embd": n_embd,
        "n_head": n_head,
        "n_layer": n_layer,
        "dropout": dropout
    }
    return config_dict


def store_config(path, args=pass_args()):
    config_dict = variables_to_dict()
    with open(path, "w") as config_file:
        config_data = {"config_args": args, "config_dict": config_dict}
        json.dump(config_data, config_file)


def retrieve_config(path):
    try:
        with open(path, "r") as config_file:
            config_data = json.load(config_file)
        return config_data
    except FileNotFoundError:
        print("warning : config.json not found using default args")
        config_data = {"config_args": pass_args()}
        return config_data


get_args()



import os
from setuptools import setup, find_packages

setup(
    name='jam-gpt',
    version='0.0.3',
    description='A reimplementation of large language model (LLM) architectures designed for research and development processes',
    author='Lokeshwaran M',
    author_email='lokeshwaran.m23072003@gmail.com',
    url="https://github.com/Lokeshwaran-M/jam-gpt.git",
    license="MIT",
    packages=find_packages(),
    package_data={'': ['requirements.txt', 'README.md']},
    install_requires=open('requirements.txt').readlines(),
    keywords='jam-gpt Jam-AI Jam-AGI',
)


# install_requires=["setuptools==67.8.0","torch==2.0.1","tiktoken"]



import os
import pickle
import torch
from . import config


torch.manual_seed(1337)


class Model:
    """ class model to work on lm models """

    def __init__(self):
        [self.vocab_size, self.batch_size, self.block_size, self.max_iters, self.eval_interval, self.learning_rate,
            self.device, self.eval_iters, self.n_embd, self.n_head, self.n_layer, self.dropout] = config.pass_args()

        self.model = None

        self.train_data = None
        self.test_data = None

    def set_parameters(self, args: list):
        [self.vocab_size, self.batch_size, self.block_size, self.max_iters, self.eval_interval, self.learning_rate,
            self.device, self.eval_iters, self.n_embd, self.n_head, self.n_layer, self.dropout] = args

    def get_parameters(self):
        return [self.vocab_size, self.batch_size, self.block_size, self.max_iters, self.eval_interval, self.learning_rate, self.device, self.eval_iters, self.n_embd, self.n_head, self.n_layer, self.dropout]

    def set_model(self, model):
        self.model = model
        self.m = self.model.to(self.device)
        # print the number of parameters in the model
        print("vocab size : ", self.vocab_size)
        print("parameters : ", sum(p.numel()
              for p in self.m.parameters())/1e6, " M")
        return self.m

    def set_data(self, data):
        self.train_data = data[0]
        self.test_data = data[1]

    def get_batch(self, split):
        # generate small batch of data of input -> x and targets -> y
        data = self.train_data if split == 'train' else self.test_data
        ix = torch.randint(len(data) - self.block_size, (self.batch_size,))
        x = torch.stack([data[i:i+self.block_size] for i in ix])
        y = torch.stack([data[i+1:i+self.block_size+1] for i in ix])
        x, y = x.to(self.device), y.to(self.device)
        return x, y

    @torch.no_grad()
    def estimate_loss(self):
        # estimates the loss of model by eval using test data
        out = {}
        self.model.eval()
        for split in ['train', 'val']:
            losses = torch.zeros(self.eval_iters)
            for k in range(self.eval_iters):
                X, Y = self.get_batch(split)
                logits, loss = self.model(X, Y)
                losses[k] = loss.item()
            out[split] = losses.mean()
        self.model.train()
        return out

    def optimize(self):
        self.optimizer = torch.optim.AdamW(
            self.model.parameters(), lr=self.learning_rate)

    def train(self, max_iters=None):
        if not max_iters:
            max_iters = self.max_iters
        for iter in range(max_iters):

            # every once in a while evaluate the loss on train and val sets
            if iter % self.eval_interval == 0:
                losses = self.estimate_loss()
                print(
                    f"step {iter}: train loss {losses['train']:.4f}, val loss {losses['val']:.4f}")

            # sample a batch of data
            xb, yb = self.get_batch('train')

            # evaluate the loss
            logits, loss = self.model(xb, yb)
            self.optimizer.zero_grad(set_to_none=True)
            loss.backward()
            self.optimizer.step()

    def generate(self, prompt, max_new_tokens=500):
        # generate from the model
        # context = torch.zeros((1, 1), dtype=torch.long, device=self.device)
        tensor_prompt = (torch.tensor(
            prompt, dtype=torch.long, device=self.device)[None, ...])

        return self.m.generate(tensor_prompt, max_new_tokens)[0].tolist()

    def save_model(self, model_name, model_format="bin"):

        # to save model
        if not os.path.exists(f"./{model_name}"):
            os.makedirs(f"./{model_name}")
        path = f"{model_name}/{model_name}.{model_format}"
        if model_format == "bin" or model_format == "pt":
            torch.save(self.model.state_dict(), path)
        elif model_format == "pkl":
            with open(path, 'wb') as f:
                pickle.dump(self.model, f)
        else:
            print(f"given model format : {model_format} is not supported")

        # to save config info
        config.set_args(self.get_parameters())
        config.store_config(f"{model_name}/config.json")

    def load_model(self, model_name, model_format="bin"):
        # to load model
        path = f"{model_name}/{model_name}.{model_format}"
        if model_format == "bin" or model_format == "pt":
            self.model.load_state_dict(torch.load(path))
        elif model_format == "pkl":
            with open(path, 'rb') as f:
                self.model = pickle.load(f)

        # to load config info
        config_data = config.retrieve_config(f"{model_name}/config.json")
        self.set_parameters(config_data["config_args"])
        config.set_args(config_data["config_args"])

        self.model.eval()
        # return self.model
